package com.kiwisoft.proguard;

import java.io.*;
import java.util.*;

import com.intellij.openapi.components.ApplicationComponent;
import com.intellij.openapi.project.Project;
import com.intellij.unscramble.UnscrambleSupport;

public class ProGuardUnscrambleSupport implements UnscrambleSupport, ApplicationComponent
{
	private String obfuscatedExceptionClassName;
	private String originalExceptionClassName;
	private String exceptionPrefix="";
	private String exceptionSuffix="";

	public String getComponentName()
	{
		return "ProGuardUnscrambleSupport";
	}

	public String getPresentableName()
	{
		return "ProGuard Unscramble Plugin";
	}

	public void initComponent()
	{
	}

	public void disposeComponent()
	{
	}

	public String unscramble(Project project, String text, String logName)
	{
		try
		{
			List stackTrace=readStackTrace(text);
			resolveStackTrace(stackTrace, logName);
			StringWriter stringWriter=new StringWriter();
			printStackTrace(stackTrace, new PrintWriter(stringWriter));
			return stringWriter.toString();
		}
		catch (Exception ex)
		{
			ex.printStackTrace();
			return null;
		}
	}

	private List readStackTrace(String text) throws IOException
	{
		List stackTrace=new ArrayList();
		Reader reader=new StringReader(text);
		LineNumberReader lineNumberReader=new LineNumberReader(reader);
		// Read the exception class name.
		while (true)
		{
			String line=lineNumberReader.readLine();
			if (line==null) throw new IOException("Can't find exception in stack trace");
			line=line.trim();
			if (line.length()==0) continue;

			if (line.startsWith("Exception in thread \""))
			{
				int quote_index=line.indexOf('"', 21);
				if (quote_index<0)
				{
					continue;
				}

				exceptionPrefix=line.substring(0, quote_index+1)+" ";
				line=line.substring(quote_index+1).trim();
			}

			int colonIndex=line.indexOf(':');
			if (colonIndex>= 0)
			{
				exceptionSuffix=line.substring(colonIndex);
				line=line.substring(0, colonIndex).trim();
			}

			obfuscatedExceptionClassName=line;

			break;
		}

		// Read the subsequent stack trace items.
		while (true)
		{
			String line=lineNumberReader.readLine();
			if (line==null)
			{
				break;
			}

			line=line.trim();

			// See if we can parse "at ___.___(___:___)", containing
			// the class name, the method name, the source file, and the
			// optional line number.
			// Just skip the line if some essential token is missing.

			if (!line.startsWith("at "))
			{
				continue;
			}

			int openParenthesisIndex=line.indexOf('(', 3);
			if (openParenthesisIndex<0)
			{
				continue;
			}

			int colonIndex=line.indexOf(':', openParenthesisIndex+1);

			int closeParenthesisIndex=line.indexOf(')', Math.max(openParenthesisIndex, colonIndex)+1);
			if (closeParenthesisIndex<0)
			{
				continue;
			}

			int periodIndex=line.lastIndexOf('.', openParenthesisIndex-1);
			if (periodIndex<0)
			{
				continue;
			}

			String obfuscatedClassName=line.substring(3, periodIndex).trim();
			String obfuscatedMethodName=line.substring(periodIndex+1, openParenthesisIndex).trim();
			String sourceFile=line.substring(openParenthesisIndex+1, colonIndex<0 ? closeParenthesisIndex : colonIndex).trim();
			int lineNumber=colonIndex<0 ? 0 : Integer.parseInt(line.substring(colonIndex+1, closeParenthesisIndex).trim());

			stackTrace.add(new MyStackTraceItem(obfuscatedClassName, obfuscatedMethodName, sourceFile, lineNumber));
		}
		return stackTrace;
	}


	/**
	 * Reads the mapping file, de-obfuscating the stack trace along the way.
	 */
	private void resolveStackTrace(List stackTrace, String logName) throws IOException
	{
		// First put the stack trace items in easily accessible structures.
		Map obfuscatedMethods=new HashMap();

		for (int index=0; index<stackTrace.size(); index++)
		{
			MyStackTraceItem item=(MyStackTraceItem)stackTrace.get(index);

			// See if this class already has other methods on the stack trace.
			HashSet classStackTraceItems=
					(HashSet)obfuscatedMethods.get(item.obfuscatedClassName);

			// Create a new entry if not.
			if (classStackTraceItems==null)
			{
				classStackTraceItems=new HashSet();
				obfuscatedMethods.put(item.obfuscatedClassName, classStackTraceItems);
			}

			// Add the item to this class's set.
			classStackTraceItems.add(item);
		}



		// Read the mapping file, resolving the stack trace as we go.
		LineNumberReader reader=null;

		try
		{
			reader=new LineNumberReader(new BufferedReader(new FileReader(logName)));

			String line=reader.readLine();

			// Read the subsequent class mappings.
			while (true)
			{
				if (line==null)
				{
					break;
				}

				line=line.trim();

				// See if we can parse "___ -> ___:", containing the original
				// class name and the obfuscated class name.

				int colonIndex=line.indexOf(':');
				if (colonIndex<1)
				{
					continue;
				}

				int arrowIndex=line.indexOf("->");
				if (arrowIndex<1)
				{
					continue;
				}

				String originalClassName=line.substring(0, arrowIndex).trim();
				String obfuscatedClassName=line.substring(arrowIndex+2, colonIndex).trim();

				// Is this the obfuscated class name of the exception?
				if (obfuscatedExceptionClassName.equals(obfuscatedClassName))
				{
					// Resolve the exception class name.
					originalExceptionClassName=originalClassName;
				}

				// Is this obfuscated class name used elsewhere in the stack trace?
				HashSet classStackTraceItems=
						(HashSet)obfuscatedMethods.get(obfuscatedClassName);

				if (classStackTraceItems!=null)
				{
					// Resolve the class names in the stack trace items.
					Iterator items=classStackTraceItems.iterator();
					while (items.hasNext())
					{
						MyStackTraceItem item=(MyStackTraceItem)items.next();
						item.originalClassName=originalClassName;
					}
				}

				// Read the subsequent method mappings of this class.
				while (true)
				{
					line=reader.readLine();

					// See if we can parse "    ___:___:___ ___(___) -> ___",
					// containing the line numbers, the return type, the original
					// method name, the original method arguments, and the
					// obfuscated method name.

					if (line==null ||
							!line.startsWith("    "))
					{
						break;
					}

					// Just ignore these class members if their class isn't
					// in the stack trace anyway.
					if (classStackTraceItems==null)
					{
						continue;
					}

					line=line.trim();

					int colonIndex1=line.indexOf(':');
					int colonIndex2=line.indexOf(':', colonIndex1+1);

					int spaceIndex=line.indexOf(' ', colonIndex2+1);
					if (spaceIndex<0)
					{
						continue;
					}

					int openParenthesisIndex=line.indexOf('(', spaceIndex+1);
					if (openParenthesisIndex<0)
					{
						continue;
					}

					int closeParenthesisIndex=line.indexOf(')', openParenthesisIndex+1);
					if (closeParenthesisIndex<0)
					{
						continue;
					}

					arrowIndex=line.indexOf("->", closeParenthesisIndex+1);
					if (arrowIndex<1)
					{
						continue;
					}

					int firstLineNumber=colonIndex1<0 ?
							0 :
							Integer.parseInt(line.substring(0, colonIndex1).trim());

					int lastLineNumber=colonIndex1<0 || colonIndex2<0 ?
							0 :
							Integer.parseInt(line.substring(colonIndex1+1, colonIndex2).trim());

					// Include return type and arguments in the method name if
					// we're in verbose mode.
//					String originalMethodName=line.substring(colonIndex2+1, closeParenthesisIndex+1).trim();
					String originalMethodName=line.substring(spaceIndex+1, openParenthesisIndex).trim();
					String obfuscatedMethodName=line.substring(arrowIndex+2).trim();

					// Resolve the method names in the stack trace items.
					Iterator items=classStackTraceItems.iterator();
					while (items.hasNext())
					{
						MyStackTraceItem item=(MyStackTraceItem)items.next();
						if (!item.uniqueSolution &&
								item.obfuscatedMethodName.equals(obfuscatedMethodName) &&
								(item.lineNumber==0 ||
								firstLineNumber==0 ||
								lastLineNumber==0 ||
								(firstLineNumber<=item.lineNumber &&
								lastLineNumber>= item.lineNumber)))
						{
							// Create a List for storing solutions for this
							// method name.
							if (item.originalMethodNames==null)
							{
								item.originalMethodNames=new ArrayList();
							}

							// Does the method have line numbers?
							if (item.lineNumber!=0 &&
									firstLineNumber!=0 &&
									lastLineNumber!=0)
							{
								// Then it will be the one and only solution.
								item.uniqueSolution=true;
								item.originalMethodNames.clear();
							}

							// Add this method name solution to the list.
							item.originalMethodNames.add(originalMethodName);
						}
					}
				}
			}
		}
		finally
		{
			if (reader!=null)
			{
				try
				{
					reader.close();
				}
				catch (IOException ex)
				{
				}
			}
		}
	}


	/**
	 * Prints out the de-obfuscated stack trace.
	 */
	private void printStackTrace(List stackTrace, PrintWriter writer)
	{
		// Get the original exception class name, if we found it.
		String exceptionClassName=originalExceptionClassName!=null ? originalExceptionClassName : obfuscatedExceptionClassName;

		writer.println(exceptionPrefix+exceptionClassName+exceptionSuffix);

		for (int index=0; index<stackTrace.size(); index++)
		{
			MyStackTraceItem item=(MyStackTraceItem)stackTrace.get(index);

			// Get the original class name, if we found it.
			String className=item.originalClassName!=null ? item.originalClassName : item.obfuscatedClassName;

			// Get the first original method name, if we found it.
			String methodName=item.originalMethodNames!=null ? (String)item.originalMethodNames.get(0) : item.obfuscatedMethodName;

			// Compose the source file with the line number, if any.
			String source=item.lineNumber!=0 ? item.sourceFile+":"+item.lineNumber : item.sourceFile;

			// Print out the resolved stack trace item.
			writer.println("\tat "+className+"."+methodName+"("+source+")");

			// Print out alternatives, if any.
			if (item.originalMethodNames!=null)
			{
				for (int otherMethodNameIndex=1; otherMethodNameIndex<item.originalMethodNames.size(); otherMethodNameIndex++)
				{
					String otherMethodName=(String)item.originalMethodNames.get(otherMethodNameIndex);
					writer.println("\t"+spaces(className.length()+4)+otherMethodName);
				}
			}
		}
	}

	private static final String SPACES="                                                                ";

	/**
	 * Returns a String of spaces of the given length, up to the maximum length
	 * of <code>SPACES</code>.
	 */
	private String spaces(int count)
	{
		StringBuffer text=new StringBuffer(SPACES.substring(0, Math.min(count, SPACES.length())));
		while (text.length()<count) text.append(" ");
		return text.toString();
	}


	/**
	 * This class represents an obfuscated stack trace item.
	 */
	private static class MyStackTraceItem
	{
		public String obfuscatedClassName;
		public String obfuscatedMethodName;
		public String sourceFile;
		public int lineNumber;

		public String originalClassName;
		public List originalMethodNames;
		public boolean uniqueSolution;

		public MyStackTraceItem(String obfuscatedClassName, String obfuscatedMethodName, String sourceFile, int lineNumber)
		{
			this.obfuscatedClassName=obfuscatedClassName;
			this.obfuscatedMethodName=obfuscatedMethodName;
			this.sourceFile=sourceFile;
			this.lineNumber=lineNumber;
		}
	}

	public static void main(String[] args)
	{
		ProGuardUnscrambleSupport unscrampler=new ProGuardUnscrambleSupport();
		String stackTrace="java.lang.NullPointerException\n"+
				"\tat com.net_linx.star.business.guiorder.ValueBOFactory.b(Unknown Source)\n"+
				"\tat com.net_linx.star.business.guiorder.ValueBOFactory.a(Unknown Source)\n"+
				"\tat com.net_linx.star.business.guiorder.ValueBOFactory.getAllPossibleValueBOs(Unknown Source)\n"+
				"\tat com.net_linx.star.gui.order.ValueBOQueryExecutable.getCollection(Unknown Source)\n"+
				"\tat com.net_linx.afwx.presentation.controller.AbstractCollectionQueryExecutable.fetchUnion(Unknown Source)\n"+
				"\tat com.net_linx.afwx.presentation.component.lookup.DefaultLookupModel.requestPopup(Unknown Source)\n"+
				"\tat com.net_linx.afwx.presentation.component.LookupPanel.requestLookup(Unknown Source)\n"+
				"\tat com.net_linx.afwx.presentation.table.DefaultTableLookupEditor.b(Unknown Source)\n"+
				"\tat com.net_linx.afwx.presentation.table.DefaultTableLookupEditor.a(Unknown Source)\n"+
				"\tat com.net_linx.afwx.presentation.table.ak.actionPerformed(Unknown Source)\n"+
				"\tat com.net_linx.afwx.presentation.component.i.actionPerformed(Unknown Source)\n"+
				"\tat javax.swing.AbstractButton.fireActionPerformed(Unknown Source)\n"+
				"\tat javax.swing.AbstractButton$ForwardActionEvents.actionPerformed(Unknown Source)\n"+
				"\tat javax.swing.DefaultButtonModel.fireActionPerformed(Unknown Source)\n"+
				"\tat javax.swing.DefaultButtonModel.setPressed(Unknown Source)\n"+
				"\tat javax.swing.plaf.basic.BasicButtonListener.mouseReleased(Unknown Source)\n"+
				"\tat java.awt.AWTEventMulticaster.mouseReleased(Unknown Source)\n"+
				"\tat java.awt.Component.processMouseEvent(Unknown Source)\n"+
				"\tat java.awt.Component.processEvent(Unknown Source)\n"+
				"\tat java.awt.Container.processEvent(Unknown Source)\n"+
				"\tat java.awt.Component.dispatchEventImpl(Unknown Source)\n"+
				"\tat java.awt.Container.dispatchEventImpl(Unknown Source)\n"+
				"\tat java.awt.Component.dispatchEvent(Unknown Source)\n"+
				"\tat javax.swing.plaf.basic.BasicTableUI$MouseInputHandler.repostEvent(Unknown Source)\n"+
				"\tat javax.swing.plaf.basic.BasicTableUI$MouseInputHandler.mouseReleased(Unknown Source)\n"+
				"\tat java.awt.AWTEventMulticaster.mouseReleased(Unknown Source)\n"+
				"\tat java.awt.AWTEventMulticaster.mouseReleased(Unknown Source)\n"+
				"\tat java.awt.AWTEventMulticaster.mouseReleased(Unknown Source)\n"+
				"\tat java.awt.AWTEventMulticaster.mouseReleased(Unknown Source)\n"+
				"\tat java.awt.AWTEventMulticaster.mouseReleased(Unknown Source)\n"+
				"\tat java.awt.Component.processMouseEvent(Unknown Source)\n"+
				"\tat java.awt.Component.processEvent(Unknown Source)\n"+
				"\tat java.awt.Container.processEvent(Unknown Source)\n"+
				"\tat java.awt.Component.dispatchEventImpl(Unknown Source)\n"+
				"\tat java.awt.Container.dispatchEventImpl(Unknown Source)\n"+
				"\tat java.awt.Component.dispatchEvent(Unknown Source)\n"+
				"\tat java.awt.LightweightDispatcher.retargetMouseEvent(Unknown Source)\n"+
				"\tat java.awt.LightweightDispatcher.processMouseEvent(Unknown Source)\n"+
				"\tat java.awt.LightweightDispatcher.dispatchEvent(Unknown Source)\n"+
				"\tat java.awt.Container.dispatchEventImpl(Unknown Source)\n"+
				"\tat java.awt.Window.dispatchEventImpl(Unknown Source)\n"+
				"\tat java.awt.Component.dispatchEvent(Unknown Source)\n"+
				"\tat java.awt.EventQueue.dispatchEvent(Unknown Source)\n"+
				"\tat java.awt.EventDispatchThread.pumpOneEventForHierarchy(Unknown Source)\n"+
				"\tat java.awt.EventDispatchThread.pumpEventsForHierarchy(Unknown Source)\n"+
				"\tat java.awt.EventDispatchThread.pumpEvents(Unknown Source)\n"+
				"\tat java.awt.EventDispatchThread.pumpEvents(Unknown Source)\n"+
				"\tat java.awt.EventDispatchThread.run(Unknown Source)";
		System.err.println((unscrampler.unscramble(null, stackTrace, "C:\\java\\projects\\plugins\\proguard\\example\\example.log")));
	}
}
